import os, sys, pytest, pandas, random
from pathlib import Path
from datetime import datetime
from pprint import pprint
from dotenv import load_dotenv

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtexceptions import *  # only contains exceptions prefixed with "SXT"

ENV = Path(Path(__file__).parents[1] / '.env')
load_dotenv(ENV, override=True)

API_URL = os.getenv('API_URL')
USER_API_KEY = os.getenv('USER_API_KEY')
SXT_STATE_BISCUIT = os.getenv('SXT_STATE_BISCUIT')    
SXTLABS_BISCUIT   = os.getenv('SXTLABS_BISCUIT')  
SXT_TELEM_BISCUIT = os.getenv('SXT_TELEM_BISCUIT')    

 

def _est_telem_queries_in_schemas():
    schemas = ['sui','sxtlabs','ethereum']
    days = 14

    # authenticate with API Key, throwing env file location off:
    sxt = SpaceAndTime()
    sxt.user.private_key = ''
    sxt.user.public_key = ''
    sxt.user.user_id = ''
    
    sxt.user.api_key = USER_API_KEY
    sxt.authenticate()

    assert sxt.user.user_id == 'pySDK_tester'
    assert not sxt.user.access_expired
    assert sxt.user.exists
    assert not sxt.user.is_trial
    assert not sxt.user.is_quota_exceeded
    assert not sxt.user.is_restricted
    assert len(sxt.access_token) > 0
    assert len(sxt.user.subscription_id) > 0
    assert sxt.user.public_key == ''
    assert sxt.user.private_key == '' 

    # run thru schema queries
    all_queries = []
    for schema in schemas:
        sxt.logger.info(f'\n\nProcessing {schema} for the last {days} days\n{"-"*30}')
        
        schema_queries = []
        sql = f"""
        Select 
          '{schema.strip()}' as Schema_Name 
        , coalesce(a.Account_Name,'Not Registered') as Sub_Name
        , count(distinct c.Subscription_ID) as Sub_Count
        , count(distinct c.USER_ID) as User_Count
        , count(*) as Queries
        , max(timestamp) as max_timestamp
        , min(timestamp) as min_timestamp
        , max(c.USER_ID) as max_UserID
        , min(c.USER_ID) as min_UserID
        FROM SXT_STATE.QUERY_META_CORE as c 
        left outer join SXTLabs.CRM_Accounts as a
        on c.Subscription_ID = a.Subscription_ID
        WHERE c.SQL_Text ilike '%{ schema.strip() }.%'
          and cast(timestamp as date) > current_date - {days}
          and c.USER_ID not in('pySDK_tester','stephen')
        group by 1,2 order by 2 desc
        """
        success, schema_queries = sxt.execute_query(sql, biscuits=[SXTLABS_BISCUIT, SXT_STATE_BISCUIT] )
        assert success

        all_queries = all_queries + schema_queries

    assert len(all_queries) == len(schemas)
    pprint(all_queries)
    pass 

    

if __name__ == "__main__":
    # test_telem_queries_in_schemas()
    pass


import logging, random, sys, json
import pandas as pd 
from io import StringIO
from datetime import datetime
from pathlib import Path

# done fighting with this, sorry
sxtpypath = str(Path(__file__).parent.resolve())
if sxtpypath not in sys.path: sys.path.append(sxtpypath)
from sxtuser import SXTUser
from sxtresource import SXTTable, SXTView
from sxtkeymanager import SXTKeyManager
from sxtenums import *
from sxtexceptions import *

class SpaceAndTime:

    user: SXTUser = None 
    application_name: str = 'SxT-SDK'
    network_calls_enabled:bool = True
    default_local_folder:str = None
    envfile_filepath:str = None
    start_time: datetime = None
    key_manager: SXTKeyManager = None
    GRANT = SXTPermission
    ENCODINGS = SXTKeyEncodings
    SQLTYPE = SXTSqlType
    OUTPUT_FORMAT = SXTOutputFormat
    TABLE_ACCESS = SXTTableAccessType
    DISCOVERY_SCOPE = SXTDiscoveryScope


    def __init__(self, envfile_filepath=None, api_url=None, 
                user_id=None, user_private_key=None, 
                default_local_folder:str = None,
                application_name='SxT-SDK', 
                logger: logging.Logger = None,
                api_key:str = None,
                authenticate:bool = False):
        """Create new instance of Space and Time SDK for Python"""
        if logger: 
            self.logger = logger 
        else: 
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
            frmt = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s','%Y-%m-%d_%H:%M:%S')
            self.logger.sxtformat = frmt
            if len(self.logger.handlers) == 0: 
                self.logger.addHandler( logging.StreamHandler() )
                self.logger.handlers[0].formatter = frmt 

        self.start_time = datetime.now()
        self.logger.info('-'*30 + f'\nSpace and Time SDK initiated for {self.application_name} at {self.start_time.strftime("%Y-%m-%d %H:%M:%S")}')

        if application_name: self.application_name = application_name
        self.default_local_folder = default_local_folder if default_local_folder else Path('.').resolve()
        self.envfile_filepath = envfile_filepath if envfile_filepath else self.default_local_folder

        self.user = SXTUser(dotenv_file=envfile_filepath, api_url=api_url, user_id=user_id, user_private_key=user_private_key, logger=self.logger, api_key=api_key)
        self.key_manager = self.user.key_manager

        if authenticate: self.authenticate()
        return None 
    
    @property
    def access_token(self) -> str:
        return self.user.access_token

    @property
    def refresh_token(self) -> str:
        return self.user.refresh_token
    
    @property
    def api_key(self) -> str:
        return self.user.api_key
    @api_key.setter
    def api_key(self, value:str):
        self.user.api_key = value

    def logger_addFileHandler(self, file:Path) -> None:
        """Adds a logging file (handler) location to the default logging object, creating any needed folders and replacing {datetime}, {date}, or {time} with sxt start_time."""
        file = Path( self.__replaceall(str(Path(file).resolve()), replacemap={}) )
        file.parent.mkdir(parents=True, exist_ok=True)
        fh = logging.FileHandler(file)
        fh.formatter = self.logger.sxtformat
        self.logger.addHandler(fh)


    def authenticate(self, user:SXTUser = None):
        """--------------------
        Authenticate user to Space and Time.  Uses the default dotenv file to create a default user, if no other is supplied.
        
        Args:
            user (SXTUser): (optional) SXTUser object used to authenticate, and set as default user.  Creates new from default dotenv file if omitted.

        Returns: 
            bool: Success indicator
            str: Access Token returned from Space and Time network

        Examples:
            >>> sxt = spaceandtime()
            >>> success, access_token = sxt.authenticate()
            >>> print( success )
            True
            >>> print( len(access_token) >= 64 )
            True

        """
        if not user: user = self.user
        if self.network_calls_enabled: 
            success, rtn = user.authenticate()
        else:
            user.access_token = 'eyJ0eXBlI_this_is_a_pretend_access_token_it_will_not_really_work_4lXUgI5gIdk8T5Rb4Zlx8-Z1rlY-0y4pu5b4lIjh60wQY_g0vkteuQE0Or0cPDbstDnLg8uRpz5dM4GNg7QHYQ'
            user.refresh_token = 'eyJ0eXBlI_this_is_a_pretend_refresh_token_it_will_not_really_work_4lXUgI5gIdk8T5Rb4Zlx8-Z1rlY-0y4pu5b4lIjh60wQY_g0vkteuQE0Or0cPDbstDnLg8uRpz5dM4GNg7QHYQ'
            success, rtn = (True, user.access_token)
        user.base_api.access_token = self.user.access_token
        self.logger.info(f'Authentication Success: {success}')
        if not success: self.logger.error(f'Authentication error: {str(rtn)}')
        return success, rtn
    

    def execute_query(self, sql_text:str, sql_type:SXTSqlType = SXTSqlType.DQL, 
                      resources:list = None, user:SXTUser = None, 
                      biscuits:list  = None, output_format:SXTOutputFormat = SXTOutputFormat.JSON) -> tuple:
        """--------------------
        Execute a query using an authenticated user.  If not specified, uses the default user.  
        
        Args: 
            sql_text (str): SQL query text to execute. Allowed two placeholders: {public_key} which will be replaced with the user.public_key, and {resource} which is replaced with the first element in resource list (resource[0]). 
            resources (list): (optional) List of Resources ("schema.table_name") in the sql_text. Supplying will optimize performance. If only 1 value, can optionally supply a str.
            sql_type (SXTSqlType): (optional) Type of query, DML, DDL, DQL. Supplying will optimize performance.
            user (SXTUser): (optional) Authenticated user to use to execute the query. Defaults to default user.
            biscuits (list): (optional) List of biscuit tokens for permissioned tables.  If only querying public tables, this is not needed.
            output_format (SXTOutputFormat): (optional) Output format enum, either JSON or CSV. Defaults to SXTOutputFormat.JSON.

        Returns:
            bool: True if success, False if in Error. 
            list: Rows, either in JSON or CSV format. 

        Examples:
            >>> from spacenadtime import SpaceAndTime
            >>> sxt = SpaceAndTime()
            >>> sxt.authenticate()
            >>> execute_query('Select 1 as A from SXTDEMO.Singularity')
            1

        """
        if not user: user = self.user
        if not resources: resources = []
        if not biscuits: biscuits = []
        rtn = []

        try: 
            resources = resources if type(resources)==list else [str(resources)]
            sql_text = self.__replaceall(mainstr=sql_text, replacemap={'resource':resources[0] if resources else [] ,'public_key':user.public_key })
            self.logger.info(f'Executing query: \n{sql_text}')

            if self.network_calls_enabled: 
                if  sql_type == SXTSqlType.DDL :
                    success, rtn = user.base_api.sql_ddl(sql_text=sql_text, biscuits=biscuits, app_name=self.application_name)

                elif sql_type == SXTSqlType.DML and resources:
                    success, rtn = user.base_api.sql_dml(sql_text=sql_text, biscuits=biscuits, app_name=self.application_name, resources=resources)

                elif sql_type == SXTSqlType.DQL and resources:
                    success, rtn = user.base_api.sql_dql(sql_text=sql_text, biscuits=biscuits, app_name=self.application_name, resources=resources)

                else:
                    success, rtn = user.base_api.sql_exec(sql_text=sql_text, biscuits=biscuits, app_name=self.application_name)
            else:
                success, rtn = (True, [{'col1':'data', 'col2':'data'},{'col1':'data', 'col2':'data'},{'col1':'data', 'col2':'data'}] )

            if not success: raise SxTQueryError(f'Query Failed: {str(rtn)}', logger=self.logger)

        except SxTQueryError as ex:
            self.logger.error(f'Error in query execution: {ex}')
            return False, {'error':f'Error in query execution: {ex}'}

        if output_format == SXTOutputFormat.JSON: return True, rtn
        if output_format == SXTOutputFormat.CSV: return self.json_to_csv(rtn)
        if output_format == SXTOutputFormat.DATAFRAME: return self.json_to_dataframe(rtn)
        if output_format == SXTOutputFormat.PARQUET: return self.json_to_parquet(rtn)
        return True, rtn
        

    def json_to_csv(self, list_of_dicts:list) -> list:
        """--------------------
        Takes a list of dictionaries (default return from DQL query) and transforms to a list of CSV rows, preceded with a header row.

        Args:
            list_of_dicts (list): A list of dictionary items, i.e., rows of JSON columns.

        Returns: 
            bool: success flag
            list: A list of CSV strings, i.e., rows of CSV values plus a header row (len(list) will always be N+1)
        """
        if list_of_dicts == []: return False, []
        try:
            rows = [','.join( list(list_of_dicts[0].keys()) )] # headers
            for row in list_of_dicts:
                rows.append( ','.join([f'"{str(val).replace(chr(34),chr(34)+chr(34))}"' for val in list(row.values())]) )
            self.logger.debug('Query JSON transformed to CSV')
            return True, rows 
        except Exception as ex:
            self.logger.error(f'Query JSON could not be transformed to CSV: {ex}')
            return False, None
            

    def json_to_dataframe(self, list_of_dicts:list) -> pd.DataFrame:
        """--------------------
        Takes a list of dictionaries (default return from DQL query) and transforms to a dataframe object.

        Args:
            list_of_dicts (list): A list of dictionary items, i.e., rows of JSON columns.

        Returns: 
            bool: success flag
            list: pandas dataframe object.
        """
        try:
            df = pd.read_json( StringIO(json.dumps(list_of_dicts)) )
            self.logger.debug('Query JSON transformed to DataFrame')
            return True, df 
        except Exception as ex:
            self.logger.error(f'Query JSON could not be transformed to DataFrame: {ex}')
            return False, None 


    def json_to_parquet(self, list_of_dicts:list) -> bytes:
        """--------------------
        Takes a list of dictionaries (default return from DQL query) and transforms to a parquet byte array.

        Args:
            list_of_dicts (list): A list of dictionary items, i.e., rows of JSON columns.

        Returns: 
            bool: success flag
            list: parquet formatted binary.
        """
        success, df = self.json_to_dataframe(list_of_dicts)
        if not success: 
            self.logger.warning('Query JSON return could not be turned into a DataFrame, and hence, not into a Parquet Binary')
            return False, None 
        try:
            pq = df.to_parquet()
            self.logger.debug('Query JSON transformed to Parquet Binary')
            return True, pq 
        except Exception as ex:
            self.logger.error(f'Query JSON could not be transformed to Parquet Binary: {ex}')
            return False, None 
            

    def __replaceall(self, mainstr:str, replacemap:dict) -> str:
        if 'date' not in replacemap.keys(): replacemap['date'] = datetime.now().strftime('%Y%m%d')
        if 'time' not in replacemap.keys(): replacemap['time'] = datetime.now().strftime('%H%M%S')
        if 'datetime' not in replacemap.keys(): replacemap['datetime'] = datetime.now().strftime('%Y%m%d_%H%M%S')
        for findname, replaceval in replacemap.items():
            mainstr = mainstr.replace('{'+str(findname)+'}', str(replaceval))                    
        return mainstr

    
    def discovery_get_schemas(self, scope:SXTDiscoveryScope = SXTDiscoveryScope.ALL, 
                              user:SXTUser = None, 
                              return_as:type = list) -> tuple:
        """--------------------
        Connects to the Space and Time network and returns all available schemas.

        Args:
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, Subscription, or Private. Defaults to SXTDiscoveryScope.ALL.
            user (SXTUser): (optional) Authenticated User object. Uses default user if omitted.
            return_as (type): (optional) Python type to return. Currently supports json, dict, list, str.

        Returns: 
            object: Return type defined with the return_as feature.
        """
        if not user: user = self.user
        if not scope: scope = SXTDiscoveryScope.ALL
        success, response = user.base_api.discovery_get_schemas(scope=scope.name)  
        if not success: 
            self.logger.warning("WARNING: base_api.discovery_get_schemas() failed to return Success")
            return False, None

        if return_as in [list, str]: 
            response = sorted([s['schema'] for s in response])
            if return_as == str: response = ', '.join(response)
        else: 
            # all other options are flavors of a dict:
            response = {s['schema']:s for s in response}

        # which flavor?
        if   return_as in [json]: response = json.dumps(response, indent=4)
        elif return_as in [dict, list, str]: pass # good as-is 
        else:
            self.logger.warning('Supplied an unsupported return type, only [json, dict, list, str] currently supported. Defaulting to dict.')
        return success, response

        
    def discovery_get_tables(self, schema:str, 
                             scope:SXTDiscoveryScope = SXTDiscoveryScope.ALL, 
                             user:SXTUser = None, 
                             search_pattern:str = None, 
                             return_as:type = json) -> tuple:
        """--------------------
        Connects to the Space and Time network and returns all available tables within a schema.

        Args:
            schema (str): Schema name to search for tables.
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, Subscription, or Private. Defaults to SXTDiscoveryScope.ALL.
            user (SXTUser): (optional) Authenticated User object. Uses default user if omitted.
            search_pattern (str): (optional) Tablename pattern to match for inclusion into result set. Defaults to None / all tables.
            return_as (type): (optional) Python type to return. Currently supports json, dict, list, str.

        Returns: 
            object: Return type defined with the return_as feature.
        """        
        if not user: user = self.user
        if not scope: scope = SXTDiscoveryScope.ALL
        success, response = user.base_api.discovery_get_tables(scope=scope.name, schema=schema, search_pattern=search_pattern)  
        if not success: 
            self.logger.warning("WARNING: base_api.discovery_get_tables() failed to return Success")
            return False, None

        if return_as in [list, str]: 
            response = sorted([ f"{r['schema']}.{r['table']}" for r in response])
            if return_as == str: response = ', '.join(response)
        else: 
            # all other options are flavors of a dict:
            response = {f"{r['schema']}.{r['table']}":r for r in response}

        # which flavor?
        if   return_as in [json]: response = json.dumps(response, indent=4)
        elif return_as in [dict, list, str]: pass # good as-is 
        else:
            self.logger.warning('Supplied an unsupported return type, only [json, dict, list, str] currently supported. Defaulting to dict.')
        return success, response



    def discovery_get_table_columns(self, schema:str, tablename:str, 
                             user:SXTUser = None, 
                             search_pattern:str = None, 
                             return_as:type = dict) -> tuple:
        """--------------------
        Connects to the Space and Time network and returns all available columns within a table.

        Args:
            schema (str): Schema name containing the below tablename.
            tablename (str): Name of table to search metadata for, and return list of column information.
            user (SXTUser): (optional) Authenticated User object. Uses default user if omitted.
            search_pattern (str): (optional) Tablename pattern to match for inclusion into result set. Defaults to None / all columns.
            return_as (type): (optional) Python type to return. Currently supports n, dict, list, str.

        Returns: 
            object: Return type defined with the return_as feature.
        """        
        if not user: user = self.user
        success, response = user.base_api.discovery_get_columns(schema=schema, table=tablename)  
        if not success: 
            self.logger.warning("WARNING: base_api.discovery_get_columns() failed to return Success")
            return False, None
            # raise SxTAPINotSuccessfulError("base_api.discovery_get_columns() failed to return Success")
        if search_pattern: response = [r for r in response if str(search_pattern).lower() in r['column'].lower()]

        # sort by 'position'
        response = sorted(response, key=lambda d: d['position'])

        if return_as in [list, str]: 
            response = ([ f"{r['column']}" for r in response])
            if return_as == str: response = ', '.join(response)
        else: 
            # all other options are flavors of a dict:
            response = {r['column']:{n:v for n,v in r.items() if n!='column'} for r in response}

        # which flavor?
        if   return_as in [json]: response = json.dumps(response, indent=4)
        elif return_as in [dict, list, str]: pass # good as-is 
        else:
            self.logger.warning('Supplied an unsupported return type, only [json, dict, list, str] currently supported. Defaulting to dict.')
        return success, response


        
    def discovery_get_views(self, schema:str, 
                             scope:SXTDiscoveryScope = SXTDiscoveryScope.ALL, 
                             user:SXTUser = None, 
                             search_pattern:str = None, 
                             return_as:type = json) -> tuple:
        """--------------------
        Connects to the Space and Time network and returns all available views within a schema, along with view text and other metadata.

        Args:
            schema (str): Schema name to search for views.
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, or Subscription. Defaults to SXTDiscoveryScope.ALL.
            user (SXTUser): (optional) Authenticated User object. Uses default user if omitted.
            search_pattern (str): (optional) Tablename pattern to match for inclusion into result set. Defaults to None / all tables.
            return_as (type): (optional) Python type to return. Currently supports json, dict (unabridged), list, str (abridged). 

        Returns: 
            object: Return type defined with the return_as feature.
        """        
        if not user: user = self.user
        if not scope: scope = SXTDiscoveryScope.ALL
        success, response = user.base_api.discovery_get_views(schema=schema, scope=scope.name, search_pattern=search_pattern)  
        if not success: 
            self.logger.warning("WARNING: base_api.discovery_get_views() failed to return Success")
            return False, None

        if return_as in [list, str]: 
            response = sorted([ f"{r['schema']}.{r['view']}" for r in response])
            if return_as == str: response = ', '.join(response)
        else:
            # all other options are flavors of a dict:
            response = {f"{r['schema']}.{r['view']}":r for r in response}

        # which flavor?
        if   return_as in [json]: response = json.dumps(response, indent=4)
        elif return_as in [dict,list, str]: pass # good as-is
        else:
            self.logger.warning('Supplied an unsupported return type, only [json, dict, list, str] currently supported. Defaulting to dict.')
        return success, response



import os, sys, pytest, pandas, random
from pathlib import Path
from datetime import datetime
from pprint import pprint
from dotenv import load_dotenv

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtexceptions import *  # only contains exceptions prefixed with "SXT"




km = SXTKeyManager(new_keypair=True)
my_private_key = km.private_key


from spaceandtime import SXTBiscuit, SXTTable

mytable = SXTTable('schema.myTable', private_key=my_private_key)
mytable.add_biscuit('Admin', SXTBiscuit.GRANT.ALL)





import os, sys, pytest, pandas, random, json
from pathlib import Path
from datetime import datetime

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtexceptions import *  # only contains exceptions prefixed with "SXT"
API_URL = 'https://api.spaceandtime.dev'

def setup_debug_logger():
    import logging
    logfile = Path(Path(__file__).resolve().parent / 'logs'/ f"pytest_debug_{datetime.now().strftime('%Y-%m-%d_%H%M%S')}.log")
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    if len(logger.handlers) == 0: 
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        formatter.default_time_format = '%Y-%m-%d %H:%M:%S'
        formatter.default_msec_format = '%s.%03d'
        # file handler:
        file_handler = logging.FileHandler(logfile)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        # console handler
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    return logger

mylogger = setup_debug_logger()

def test_sxt_addfilehandler():
    logfilepath =  Path(Path(__file__).parent / 'logs' / 'test_sxt_addfilehandler.log')
    sxt = SpaceAndTime()

    for test in [logfilepath, str(logfilepath)]:
        logfilepath.unlink(missing_ok=True)
        assert not logfilepath.exists()
        sxt.logger_addFileHandler(test)
        sxt.logger.info('test message')
        assert logfilepath.exists()
        sxt.logger.handlers.clear()



def test_sxt_exceptions():
    mylogger.info(f'\n\ntest_sxt_exceptions\n{"-"*30}')
    # test common exceptions
    sxt = SpaceAndTime() 
    sxt.user.user_id = sxt.user.api_key = ''
    with pytest.raises(Exception) as e_info: sxt.authenticate()

    sxt = SpaceAndTime() 
    sxt.user.private_key = sxt.user.api_key = ''
    with pytest.raises(Exception) as e_info: sxt.authenticate()

    with pytest.raises(SxTAuthenticationError) as errinfo:   raise SxTAuthenticationError('test message: SxTAuthenticationError')
    with pytest.raises(SxTQueryError) as errinfo:            raise SxTQueryError('test message: SxTQueryError')
    with pytest.raises(SxTFileContentError) as errinfo:      raise SxTFileContentError('test message: SxTFileContentError')
    with pytest.raises(SxTArgumentError) as errinfo:         raise SxTArgumentError('test message: SxTArgumentError')
    with pytest.raises(SxTKeyEncodingError) as errinfo:      raise SxTKeyEncodingError('test message: SxTKeyEncodingError')
    with pytest.raises(SxTBiscuitError) as errinfo:          raise SxTBiscuitError('test message: SxTBiscuitError')
    with pytest.raises(SxTAPINotDefinedError) as errinfo:    raise SxTAPINotDefinedError('test message: SxTAPINotDefinedError')
    with pytest.raises(SxTAPINotSuccessfulError) as errinfo: raise SxTAPINotSuccessfulError('test message: SxTAPINotSuccessfulError')


def test_sxt_wrapper():
    mylogger.info(f'\n\ntest_sxt_wrapper\n{"-"*30}')
    # pick up default .env file, with USERID="pySDK_tester" or... testuser_977604126 ?
    # note, that specific user must exist in .env this test to succeed.
    envfile = Path(Path(__file__).parents[1] / '.env').resolve()
    sxt = SpaceAndTime(envfile_filepath = envfile, logger= setup_debug_logger() )
    sxt.user.user_id = 'pySDK_tester'
    assert sxt.user.user_id == 'pySDK_tester'
    assert sxt.user.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert sxt.user.private_key[:6] == 'MeaW6J'

    assert len(sxt.access_token) == 0
    assert sxt.user.subscription_id == None
    assert sxt.user.is_quota_exceeded == None
    assert sxt.user.is_restricted == None
    assert sxt.user.is_trial == None
    sxt.authenticate()
    assert len(sxt.access_token) > 0
    assert sxt.user.subscription_id != None
    assert sxt.user.is_quota_exceeded != None
    assert sxt.user.is_restricted != None
    assert sxt.user.is_trial != None
    assert sxt.access_token[:4] == 'eyJ0' 
    assert sxt.user.user_id == 'pySDK_tester'

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert type(data) == list
    assert type(data[0]) == dict
  
    
def test_sxt_user():
    mylogger.info(f'\n\ntest_sxt_user\n{"-"*30}')
    # pick up default .env file, with USERID="pySDK_tester"
    # note, that specific user must be used for this test to succeed.
    print('test_sxt_user')
    sxt = None

    # UserA -- load .env file
    userA = SXTUser(dotenv_file='./.env', api_url=API_URL, logger= setup_debug_logger() )
    assert userA.user_id == 'pySDK_tester'
    assert userA.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert userA.private_key[:6] == 'MeaW6J'

    assert len(userA.access_token) == 0
    userA.authenticate()
    assert len(userA.access_token) > 0
    assert userA.access_token[:4] == 'eyJ0' 

    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'


    # UserB -- load alternate .env file
    userB = SXTUser(dotenv_file='./.env_alt', api_url=API_URL, authenticate=True, logger= setup_debug_logger() )
    # assert userB.user_id == 'pySDK_tester2'
    # assert userB.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    # assert userB.private_key[:6] == 'MeaW6J'
    assert userB.user_id == 'stephen_cli'
    assert userB.public_key == "S4HCEEe5Hlp0ePANRkNF7xrb3zasKz87H9QQ5ZcT9fU="
    assert userB.private_key[:6] == 'z0STaV'

    # authenticate flag used in initializer
    assert len(userB.access_token) > 0
    assert userB.access_token[:4] == 'eyJ0' 

    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'B'


    # Alternate querying with different users
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success

    # assert userA.user_id != userB.user_id
    # assert userA.access_token != userB.access_token

    # this is just for backwards compatibility... prefer not to use:
    success, data = userA.execute_sql("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'



def test_sxt_user_2():
    mylogger.info(f'\n\ntest_sxt_user_2\n{"-"*30}')
    # pick up default .env file, with USERID="pySDK_tester"
    # note, that specific user must be used for this test to succeed.
    print('test_sxt_user')
    sxt = None

    # UserA -- load .env file
    userA = SXTUser(dotenv_file='./.env', api_url=API_URL, logger= setup_debug_logger() )
    assert userA.user_id == 'pySDK_tester'
    assert userA.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert userA.private_key[:6] == 'MeaW6J'

    assert len(userA.access_token) == 0
    userA.authenticate()
    assert len(userA.access_token) > 0
    assert userA.access_token[:4] == 'eyJ0' 

    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'


    # UserB -- load alternate .env file
    userB = userA # SXTUser(dotenv_file='./.env_alt', api_url=API_URL, authenticate=True, logger= setup_debug_logger() )
    # assert userB.user_id == 'pySDK_tester2'
    assert userB.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert userB.private_key[:6] == 'MeaW6J'

    # authenticate flag used in initializer
    assert len(userB.access_token) > 0
    assert userB.access_token[:4] == 'eyJ0' 

    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'B'


    # Alternate querying with different users
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success

    # assert userA.user_id != userB.user_id
    # assert userA.access_token != userB.access_token

    # this is just for backwards compatibility... prefer not to use:
    success, data = userA.execute_sql("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'


 
def test_execute_query():
    mylogger.info(f'\n\ntest_execute_query\n{"-"*30}')
    sxt = SpaceAndTime()
    sxt.authenticate()
    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert type(data) == list
    assert type(data[0]) == dict

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.PARQUET )
    assert success
    assert type(data) == bytes
 
    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.DATAFRAME )
    assert success
    assert type(data) == pandas.DataFrame

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.CSV )
    assert success
    assert type(data) == list
    assert type(data[0]) == str # header
    assert type(data[1]) == str # data
    assert len(data) == 2 # header + 1 data row
    assert data[1].count(',') > 3


def test_discovery():
    mylogger.info(f'\n\ntest_discovery\n{"-"*30}')
    sxt = SpaceAndTime()
    sxt.authenticate()

    # Schemas
    success, schemas = sxt.discovery_get_schemas(return_as=list)
    assert success
    assert type(schemas) == list 
    assert 'ETHEREUM' in schemas
    assert 'SXTDEMO' in schemas
    assert 'SXTLABS' in schemas

    success, schemas = sxt.discovery_get_schemas(return_as=dict)
    assert success
    assert type(schemas) == dict

    success, schemas = sxt.discovery_get_schemas(return_as=str)
    assert success
    assert type(schemas) == str
    assert 'POLYGON,' in schemas
    assert schemas.count(',') >= 10

    success, all_schemas = sxt.discovery_get_schemas(scope = sxt.DISCOVERY_SCOPE.ALL)
    assert success
    success, sub_schemas = sxt.discovery_get_schemas(scope = sxt.DISCOVERY_SCOPE.SUBSCRIPTION)
    assert success
    assert len(all_schemas) > len(sub_schemas)

    # Tables
    success, tables = sxt.discovery_get_tables('SXTLabs', scope = sxt.DISCOVERY_SCOPE.SUBSCRIPTION, return_as=list)
    assert success
    assert 'SXTLABS.CRM_ACCOUNTS' in tables
    assert len(tables) >=10

    success, tables = sxt.discovery_get_tables('SXTLabs', search_pattern='CRM_Cosell', scope = sxt.DISCOVERY_SCOPE.SUBSCRIPTION, return_as=list)
    assert success
    assert 'SXTLABS.CRM_COSELL_AGREEMENTS' in tables
    assert len(tables) <=10

    # Columns
    success, columns = sxt.discovery_get_table_columns('POLYGON', 'BLOCKS', return_as=list)
    assert success
    assert 'TIME_STAMP' in columns
    assert len(columns) > 5
    
    success, columns = sxt.discovery_get_table_columns('POLYGON', 'BLOCKS', search_pattern='BLOCK', return_as=dict)
    assert success
    assert 'BLOCK_NUMBER' in columns.keys()
    assert len(columns) < 5
    
    success, columns = sxt.discovery_get_table_columns('SXTLABS', 'CRM_ACCOUNTS') # defaults to dict
    assert success
    assert 'CREATED_TIME' in columns.keys()
    assert 'ACCOUNT_NAME' in columns.keys()
    assert len(columns) > 15

    success, views = sxt.discovery_get_views('SXTLabs', scope = sxt.DISCOVERY_SCOPE.ALL, return_as=list)
    assert success
    assert len(views) > 0
    assert type(views[0]) == str

    success, views = sxt.discovery_get_views('SXTLabs', return_as=json)
    assert success
    assert len(views) > 0
    assert type(views) == str
    jsonreturn = json.loads(views)
    assert type(jsonreturn) == dict
    assert jsonreturn[list(jsonreturn.keys())[0]]['schema'].upper() == 'SXTLABS'


if __name__ == '__main__':
    # test_sxt_addfilehandler()
    # test_sxt_exceptions()
    # test_sxt_wrapper()
    # test_sxt_user_2()
    # test_sxt_user()
    # test_execute_query()
    # test_discovery()

    # logger = setup_debug_logger()
    # logger.info('\n\nDone!!!')
    pass 


import os, sys
from pathlib import Path
from datetime import datetime
from dotenv import load_dotenv

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtexceptions import *  # only contains exceptions prefixed with "SXT"

ENV = Path(Path(__file__).parents[1] / '.env')
load_dotenv(ENV, override=True)
API_URL = os.getenv('API_URL')
USER_API_KEY = os.getenv('USER_API_KEY')
SXT_STATE_BISCUIT = os.getenv('SXT_STATE_BISCUIT')    
SXTLABS_BISCUIT   = os.getenv('SXTLABS_BISCUIT')  
SXT_TELEM_BISCUIT = os.getenv('SXT_TELEM_BISCUIT')    


def test_apikey_login():
    
    # authenticate with API Key, throwing env file location off:
    sxt = SpaceAndTime()
    sxt.user.private_key = ''
    sxt.user.public_key = ''
    sxt.user.user_id = ''
    
    sxt.user.api_key = USER_API_KEY
    sxt.authenticate()
    
    assert sxt.user.user_id == 'pySDK_tester'
    assert not sxt.user.access_expired
    assert sxt.user.exists
    assert not sxt.user.is_trial
    assert not sxt.user.is_quota_exceeded
    assert not sxt.user.is_restricted
    assert len(sxt.access_token) > 0
    assert len(sxt.user.subscription_id) > 0
    assert sxt.user.public_key == ''
    assert sxt.user.private_key == '' 

    # execute test query
    success, data = sxt.execute_query(biscuits=[SXTLABS_BISCUIT, SXT_STATE_BISCUIT], 
                                      sql_text="""Select * from sxtlabs.singularity""")
    assert success
    assert len(data) == 1

 
if __name__ == "__main__":
    test_apikey_login()
    pass